@@ -29,17 +29,26 @@ from __future__ import absolute_import, unicode_literals, division
 
 from .hid import STATUS
 from .ctap import CtapError
-from .ctap1 import CTAP1, APDU, ApduError
-from .ctap2 import CTAP2, PinProtocolV1, AttestationObject, AssertionResponse, Info
+from .ctap1 import Ctap1, APDU, ApduError
+from .ctap2 import (
+    Ctap2,
+    AttestationObject,
+    AssertionResponse,
+    Info,
+    ClientPin,
+)
+from .ctap2.extensions import Ctap2Extension
 from .webauthn import (
     PublicKeyCredentialCreationOptions,
     PublicKeyCredentialRequestOptions,
     AuthenticatorSelectionCriteria,
     UserVerificationRequirement,
+    AuthenticatorAttestationResponse,
+    AuthenticatorAssertionResponse,
 )
 from .cose import ES256
 from .rpid import verify_rp_id, verify_app_id
-from .utils import sha256, hmac_sha256, websafe_decode, websafe_encode
+from .utils import sha256, websafe_decode, websafe_encode
 from enum import Enum, IntEnum, unique
 from threading import Timer, Event
 
@@ -183,7 +192,7 @@ class U2fClient(object):
 
     def __init__(self, device, origin, verify=verify_app_id):
         self.poll_delay = 0.25
-        self.ctap = CTAP1(device)
+        self.ctap = Ctap1(device)
         self.origin = origin
         self._verify = verify
 
@@ -191,7 +200,7 @@ class U2fClient(object):
         try:
             if self._verify(app_id, self.origin):
                 return
-        except Exception:
+        except Exception:  # nosec
             pass  # Fall through to ClientError
         raise ClientError.ERR.BAD_REQUEST()
 
@@ -264,7 +273,7 @@ class U2fClient(object):
                         key_handle,
                     )
                     break
-                except ClientError:
+                except ClientError:  # nosec
                     pass  # Ignore and try next key
         else:
             raise ClientError.ERR.DEVICE_INELIGIBLE()
@@ -291,7 +300,7 @@ class _BaseClient(object):
         try:
             if self._verify(rp_id, self.origin):
                 return
-        except Exception:
+        except Exception:  # nosec
             pass  # Fall through to ClientError
         raise ClientError.ERR.BAD_REQUEST()
 
@@ -304,6 +313,57 @@ class _BaseClient(object):
         )
 
 
+class AssertionSelection(object):
+    """GetAssertion result holding one or more assertions.
+
+    Since multiple assertions may be retured by Fido2Client.get_assertion, this result
+    is returned which can be used to select a specific response to get.
+    """
+
+    def __init__(self, client_data, assertions):
+        self._client_data = client_data
+        self._assertions = assertions
+
+    def get_assertions(self):
+        """Get the raw AssertionResponses available to inspect before selecting one."""
+        return self._assertions
+
+    def _get_extension_results(self, assertion):
+        return {}
+
+    def get_response(self, index):
+        """Get a single response."""
+        assertion = self._assertions[index]
+
+        return AuthenticatorAssertionResponse(
+            self._client_data,
+            assertion.auth_data,
+            assertion.signature,
+            assertion.user["id"] if assertion.user else None,
+            assertion.credential["id"] if assertion.credential else None,
+            self._get_extension_results(assertion),
+        )
+
+
+class Fido2ClientAssertionSelection(AssertionSelection):
+    def __init__(self, client_data, assertions, extensions):
+        super(Fido2ClientAssertionSelection, self).__init__(client_data, assertions)
+        self._extensions = extensions
+
+    def _get_extension_results(self, assertion):
+        # Process extenstion outputs
+        extension_outputs = {}
+        for ext in self._extensions:
+            output = ext.process_get_output(assertion.auth_data)
+            if output is not None:
+                extension_outputs.update(output)
+        return extension_outputs
+
+
+def _default_extensions():
+    return [cls for cls in Ctap2Extension.__subclasses__() if hasattr(cls, "NAME")]
+
+
 _CTAP1_INFO = Info.create(["U2F_V2"])
 
 
@@ -318,21 +378,22 @@ class Fido2Client(_BaseClient):
     :param verify: Function to verify an RP ID for a given origin.
     """
 
-    def __init__(self, device, origin, verify=verify_rp_id):
+    def __init__(self, device, origin, verify=verify_rp_id, extension_types=None):
         super(Fido2Client, self).__init__(origin, verify)
 
+        self.extensions = extension_types or _default_extensions()
         self.ctap1_poll_delay = 0.25
         try:
-            self.ctap2 = CTAP2(device)
-            self.info = self.ctap2.get_info()
-            if PinProtocolV1.VERSION in self.info.pin_protocols:
-                self.pin_protocol = PinProtocolV1(self.ctap2)
-            else:
-                self.pin_protocol = None
+            self.ctap2 = Ctap2(device)
+            self.info = self.ctap2.info
+            try:
+                self.client_pin = ClientPin(self.ctap2)
+            except ValueError:
+                self.client_pin = None
             self._do_make_credential = self._ctap2_make_credential
             self._do_get_assertion = self._ctap2_get_assertion
-        except ValueError:
-            self.ctap1 = CTAP1(device)
+        except (ValueError, CtapError):
+            self.ctap1 = Ctap1(device)
             self.info = _CTAP1_INFO
             self._do_make_credential = self._ctap1_make_credential
             self._do_get_assertion = self._ctap1_get_assertion
@@ -391,21 +452,21 @@ class Fido2Client(_BaseClient):
         selection = options.authenticator_selection or AuthenticatorSelectionCriteria()
 
         try:
-            return (
-                self._do_make_credential(
-                    client_data,
-                    options.rp,
-                    options.user,
-                    options.pub_key_cred_params,
-                    options.exclude_credentials,
-                    options.extensions,
-                    selection.require_resident_key,
-                    self._get_ctap_uv(selection.user_verification, pin is not None),
-                    pin,
-                    event,
-                    kwargs.get("on_keepalive"),
-                ),
+            att_obj, extension_outputs = self._do_make_credential(
                 client_data,
+                options.rp,
+                options.user,
+                options.pub_key_cred_params,
+                options.exclude_credentials,
+                options.extensions,
+                selection.require_resident_key,
+                self._get_ctap_uv(selection.user_verification, pin is not None),
+                pin,
+                event,
+                kwargs.get("on_keepalive"),
+            )
+            return AuthenticatorAttestationResponse(
+                client_data, att_obj, extension_outputs,
             )
         except CtapError as e:
             raise _ctap2client_err(e)
@@ -427,12 +488,17 @@ class Fido2Client(_BaseClient):
         event,
         on_keepalive,
     ):
+        self.info = self.ctap2.get_info()  # Make sure we have "fresh" info
         pin_auth = None
         pin_protocol = None
         if pin:
-            pin_protocol = self.pin_protocol.VERSION
-            pin_token = self.pin_protocol.get_pin_token(pin)
-            pin_auth = hmac_sha256(pin_token, client_data.hash)[:16]
+            pin_protocol = self.client_pin.protocol.VERSION
+            pin_token = self.client_pin.get_pin_token(
+                pin, ClientPin.PERMISSION.MAKE_CREDENTIAL, rp["id"]
+            )
+            pin_auth = self.client_pin.protocol.authenticate(
+                pin_token, client_data.hash
+            )
         elif self.info.options.get("clientPin") and not uv:
             raise ClientError.ERR.BAD_REQUEST("PIN required but not provided")
 
@@ -456,13 +522,23 @@ class Fido2Client(_BaseClient):
             if max_creds and len(exclude_list) > max_creds:
                 raise ClientError.ERR.BAD_REQUEST("exclude_list too long")
 
-        return self.ctap2.make_credential(
+        # Process extensions
+        client_inputs = extensions or {}
+        extension_inputs = {}
+        used_extensions = []
+        for ext in [cls(self.ctap2) for cls in self.extensions]:
+            auth_input = ext.process_create_input(client_inputs)
+            if auth_input is not None:
+                used_extensions.append(ext)
+                extension_inputs[ext.NAME] = auth_input
+
+        att_obj = self.ctap2.make_credential(
             client_data.hash,
             rp,
             user,
             key_params,
-            exclude_list,
-            extensions,
+            exclude_list or None,
+            extension_inputs or None,
             options,
             pin_auth,
             pin_protocol,
@@ -470,6 +546,15 @@ class Fido2Client(_BaseClient):
             on_keepalive,
         )
 
+        # Process extenstion outputs
+        extension_outputs = {}
+        for ext in used_extensions:
+            output = ext.process_create_output(att_obj.auth_data)
+            if output is not None:
+                extension_outputs.update(output)
+
+        return att_obj, extension_outputs
+
     def _ctap1_make_credential(
         self,
         client_data,
@@ -507,16 +592,19 @@ class Fido2Client(_BaseClient):
                     )
                     raise ClientError.ERR.DEVICE_INELIGIBLE()
 
-        return AttestationObject.from_ctap1(
-            app_param,
-            _call_polling(
-                self.ctap1_poll_delay,
-                event,
-                on_keepalive,
-                self.ctap1.register,
-                client_data.hash,
+        return (
+            AttestationObject.from_ctap1(
                 app_param,
+                _call_polling(
+                    self.ctap1_poll_delay,
+                    event,
+                    on_keepalive,
+                    self.ctap1.register,
+                    client_data.hash,
+                    app_param,
+                ),
             ),
+            {},
         )
 
     def get_assertion(self, options, **kwargs):
@@ -543,18 +631,18 @@ class Fido2Client(_BaseClient):
         )
 
         try:
-            return (
-                self._do_get_assertion(
-                    client_data,
-                    options.rp_id,
-                    options.allow_credentials,
-                    options.extensions,
-                    self._get_ctap_uv(options.user_verification, pin is not None),
-                    pin,
-                    event,
-                    kwargs.get("on_keepalive"),
-                ),
+            assertions, used_extensions = self._do_get_assertion(
                 client_data,
+                options.rp_id,
+                options.allow_credentials,
+                options.extensions,
+                self._get_ctap_uv(options.user_verification, pin is not None),
+                pin,
+                event,
+                kwargs.get("on_keepalive"),
+            )
+            return Fido2ClientAssertionSelection(
+                client_data, assertions, used_extensions,
             )
         except CtapError as e:
             raise _ctap2client_err(e)
@@ -565,12 +653,17 @@ class Fido2Client(_BaseClient):
     def _ctap2_get_assertion(
         self, client_data, rp_id, allow_list, extensions, uv, pin, event, on_keepalive
     ):
+        self.info = self.ctap2.get_info()  # Make sure we have "fresh" info
         pin_auth = None
         pin_protocol = None
         if pin:
-            pin_protocol = self.pin_protocol.VERSION
-            pin_token = self.pin_protocol.get_pin_token(pin)
-            pin_auth = hmac_sha256(pin_token, client_data.hash)[:16]
+            pin_protocol = self.client_pin.protocol.VERSION
+            pin_token = self.client_pin.get_pin_token(
+                pin, ClientPin.PERMISSION.GET_ASSERTION, rp_id
+            )
+            pin_auth = self.client_pin.protocol.authenticate(
+                pin_token, client_data.hash
+            )
         elif self.info.options.get("clientPin") and not uv:
             raise ClientError.ERR.BAD_REQUEST("PIN required but not provided")
 
@@ -592,11 +685,21 @@ class Fido2Client(_BaseClient):
             if max_creds and len(allow_list) > max_creds:
                 raise ClientError.ERR.BAD_REQUEST("allow_list too long")
 
-        return self.ctap2.get_assertions(
+        # Process extensions
+        client_inputs = extensions or {}
+        extension_inputs = {}
+        used_extensions = []
+        for ext in [cls(self.ctap2) for cls in self.extensions]:
+            auth_input = ext.process_get_input(client_inputs)
+            if auth_input is not None:
+                used_extensions.append(ext)
+                extension_inputs[ext.NAME] = auth_input
+
+        assertions = self.ctap2.get_assertions(
             rp_id,
             client_data.hash,
-            allow_list,
-            extensions,
+            allow_list or None,
+            extension_inputs or None,
             options,
             pin_auth,
             pin_protocol,
@@ -604,6 +707,8 @@ class Fido2Client(_BaseClient):
             on_keepalive,
         )
 
+        return assertions, used_extensions
+
     def _ctap1_get_assertion(
         self, client_data, rp_id, allow_list, extensions, uv, pin, event, on_keepalive
     ):
@@ -623,7 +728,8 @@ class Fido2Client(_BaseClient):
                     app_param,
                     cred["id"],
                 )
-                return [AssertionResponse.from_ctap1(app_param, cred, auth_resp)]
+                assertions = [AssertionResponse.from_ctap1(app_param, cred, auth_resp)]
+                return assertions, []
             except ClientError as e:
                 if e.code == ClientError.ERR.TIMEOUT:
                     raise  # Other errors are ignored so we move to the next.
@@ -640,7 +746,7 @@ if platform.system().lower() == "windows":
             WebAuthNUserVerificationRequirement,
             WebAuthNAttestationConvoyancePreference,
         )
-    except Exception:  # TODO: Make this less generic
+    except Exception:  # nosec # TODO: Make this less generic
         pass
 
 
@@ -712,7 +818,9 @@ class WindowsClient(_BaseClient):
         except OSError as e:
             raise ClientError.ERR.OTHER_ERROR(e)
 
-        return AttestationObject(result), client_data
+        return AuthenticatorAttestationResponse(
+            client_data, AttestationObject(result), {}
+        )
 
     def get_assertion(self, options, **kwargs):
         """Get assertion using Windows WebAuthN APIs.
@@ -746,7 +854,7 @@ class WindowsClient(_BaseClient):
             raise ClientError.ERR.OTHER_ERROR(e)
 
         user = {"id": user_id} if user_id else None
-        return (
-            [AssertionResponse.create(credential, auth_data, signature, user)],
+        return AssertionSelection(
             client_data,
+            [AssertionResponse.create(credential, auth_data, signature, user)],
         )
